/*
	
	ArpRunningBar.c
	
	Copyright (c)1998 by Angry Red Planet.

	This code is distributed under a modified form of the
	Artistic License.  A copy of this license should have
	been included with it; if this wasn't the case, the
	entire package can be obtained at
	<URL:http://www.angryredplanet.com/>.

	A layout class that organizes its children as a horizontal
	or vertical row.
*/

#ifndef ARPLAYOUT_ARPRUNNINGBAR_H
#include "ArpLayout/ArpRunningBar.h"
#endif

#ifndef ARPKERNEL_ARPDEBUG_H
#include "ArpKernel/ArpDebug.h"
#endif

#ifndef _AUTOLOCK_H
#include <be/support/Autolock.h>
#endif

#include <algobase.h>
#include <float.h>
#include <string.h>

ArpMOD();

const char* ArpRunningBar::WeightC = "Weight";
const char* ArpRunningBar::FillC = "Fill";
const char* ArpRunningBar::OrientationP = "Orientation";
const char* ArpRunningBar::InsetLeftP = "InsetLeft";
const char* ArpRunningBar::InsetRightP = "InsetRight";
const char* ArpRunningBar::InsetTopP = "InsetTop";
const char* ArpRunningBar::InsetBottomP = "InsetBottom";
const char* ArpRunningBar::IntraSpaceP = "IntraSpace";

/* ------ ArpRunningBar constructor and destructor ------
 *
 * The various ways to create and destroy ArpRunningBar
 * objects.
 */
 
ArpRunningBar::ArpRunningBar(const char* name)
		: ArpLayoutable(name)
{
	PG_Orientation =
		PG_InsetLeft = PG_InsetRight = PG_InsetTop = PG_InsetBottom =
		PG_IntraSpace = (const char*)0;
	PV_Orientation = B_HORIZONTAL;
	PV_InsetLeft = PV_InsetRight = PV_InsetTop = PV_InsetBottom =
		PV_IntraSpace = 0;
	total_weight = 0;
	char_w = char_h = 0;
	inset_l = inset_r = inset_t = inset_b = 0;
	intra_s = 0;
}

ArpRunningBar::ArpRunningBar(BMessage* data)
	: ArpLayoutable(data)
{
	/* XXX We need to initialize these BEFORE!!
	PG_Orientation =
		PG_InsetLeft = PG_InsetRight = PG_InsetTop = PG_InsetBottom =
		PG_IntraSpace = NULL;
	PV_Orientation = B_HORIZONTAL;
	PV_InsetLeft = PV_InsetRight = PV_InsetTop = PV_InsetBottom =
		PV_IntraSpace = 0;
	*/
}

/* ------------ ArpRunningBar archiving ------------
 *
 * Archiving and retrieving ArpLayoutable objects.
 */
ArpRunningBar* ArpRunningBar::Instantiate(BMessage* archive)
{
	if ( validate_instantiation(archive, "ArpRunningBar") ) 
		return new ArpRunningBar(archive); 
	return NULL;
}

status_t ArpRunningBar::Archive(BMessage* data, bool deep)
{
	status_t status = inherited::Archive(data,deep);
	if( status == B_NO_ERROR )
		status = data->AddString("class","ArpRunningBar");
	return status;
}

void ArpRunningBar::DoSetParams(const ArpMessage* g,
								const ArpMessage* p)
{
	inherited::DoSetParams(g,p);
	
	ArpD(cdb << ADH << "ArpRunningBar::DoSetParams(), with:" << endl);
	ArpD(if( g ) cdb << ADH << "Globals = " << *g << endl);
	ArpD(if( p ) cdb << ADH << "Params = " << *p << endl);
	
	status_t res;
	res = ExtractParam(p,OrientationP,B_INT32_TYPE,
						g,&PG_Orientation,&PV_Orientation);
	if( !res ) InvalidateDimens();
	res = ExtractParam(p,InsetLeftP,B_FLOAT_TYPE,
						g,&PG_InsetLeft,&PV_InsetLeft);
	if( !res ) InvalidateDimens();
	res = ExtractParam(p,InsetRightP,B_FLOAT_TYPE,
						g,&PG_InsetRight,&PV_InsetRight);
	if( !res ) InvalidateDimens();
	res = ExtractParam(p,InsetTopP,B_FLOAT_TYPE,
						g,&PG_InsetTop,&PV_InsetTop);
	if( !res ) InvalidateDimens();
	res = ExtractParam(p,InsetBottomP,B_FLOAT_TYPE,
						g,&PG_InsetBottom,&PV_InsetBottom);
	if( !res ) InvalidateDimens();
	res = ExtractParam(p,IntraSpaceP,B_FLOAT_TYPE,
						g,&PG_IntraSpace,&PV_IntraSpace);
	if( !res ) InvalidateDimens();
}

bool ArpRunningBar::DoParams(ArpMessage& p, const char* name)
{
	bool res = inherited::DoParams(p,name);
	if( !name || !strcmp(name,OrientationP) ) {
		if( p.AddInt32(OrientationP,PV_Orientation) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,InsetLeftP) ) {
		if( p.AddFloat(InsetLeftP,PV_InsetLeft) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,InsetRightP) ) {
		if( p.AddFloat(InsetRightP,PV_InsetRight) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,InsetTopP) ) {
		if( p.AddFloat(InsetTopP,PV_InsetTop) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,InsetBottomP) ) {
		if( p.AddFloat(InsetBottomP,PV_InsetBottom) ) {
			res = false;
		}
		if( name ) return res;
	}
	if( !name || !strcmp(name,IntraSpaceP) ) {
		if( p.AddFloat(IntraSpaceP,PV_IntraSpace) ) {
			res = false;
		}
		if( name ) return res;
	}
	return res;
}

void ArpRunningBar::ComputeDimens(void)
{
	int32 num = CountLayoutChildren();

	// Initialize dimensions of this component.
	cur_dimens.min_width = cur_dimens.min_height
		= cur_dimens.pref_width = cur_dimens.pref_height
		= cur_dimens.max_width = cur_dimens.max_height = 0;
	total_weight = 0;
	min_weight = 0;
	
	// Retrieve font metrics.
	char_w = PV_BasicFont.StringWidth("W");
	font_height fhs;
	PV_BasicFont.GetHeight(&fhs);
	char_h = fhs.ascent+fhs.descent+fhs.leading;
	
	ArpD(cdb << ADH << "charleft=" << PV_InsetLeft
				<< " charright=" << PV_InsetRight
				<< " chartop=" << PV_InsetTop
				<< " charbot=" << PV_InsetBottom
				<< " charspace=" << PV_IntraSpace << endl);
	
	// Compute inset pixel values based on font metrics.
	inset_l = ceil(PV_InsetLeft*char_w);
	inset_r = ceil(PV_InsetRight*char_w);
	inset_t = ceil(PV_InsetTop*char_h);
	inset_b = ceil(PV_InsetBottom*char_h);
	intra_s = ceil(PV_IntraSpace *
					((PV_Orientation==B_VERTICAL) ? char_w:char_h));
					
	ArpD(cdb << ADH << "pixeft=" << inset_l
				<< " pixright=" << inset_r
				<< " pixtop=" << inset_t
				<< " pixbot=" << inset_b
				<< " pixspace=" << intra_s << endl);
				
	for( int32 i=0; i<num; i++ ) {
		ArpLayoutable* child = LayoutChildAt(i);
		if( child ) {
			// Retrieve dimensions and weight of this child.
			const ArpLayoutDimens& dim(child->LayoutDimens());
			total_weight += child->Constraints().GetFloat(WeightC,1);
			
			// For horizontal orientation...
			if( PV_Orientation == B_HORIZONTAL ) {
			
				// The weight when our size is less than the preferred
				// size is the difference between all of the preferred
				// sizes and the minimum sizes.  This allows us to later
				// smoothly scale between the preferred size layout to
				// the minimum size layout.
				min_weight += dim.pref_width-dim.min_width;
				
				// Add intra-child space if there is more than one.
				if( i > 0 ) {
					cur_dimens.min_width += intra_s;
					cur_dimens.pref_width += intra_s;
				}
				
				// Add in running dimensions of child.  For the max size, we
				// know [because we wrote ArpLayoutable ;)] that anything
				// => ArpAnySize means "no maximum", so special case that.
				cur_dimens.min_width += dim.min_width;
				cur_dimens.pref_width += dim.pref_width;
				if( cur_dimens.max_width < ArpAnySize ) {
					cur_dimens.max_width += dim.max_width;
					if( i > 0 ) cur_dimens.max_width += intra_s;
				}
				
				// Add in lined dimensions of the child.  Note that we
				// take the max() for the maximum height as well, so that
				// we can always be scaled out to fill out the largest
				// child.
				cur_dimens.min_height =
					max(cur_dimens.min_height,dim.min_height);
				cur_dimens.pref_height =
					max(cur_dimens.pref_height,dim.pref_height);
				if( cur_dimens.max_height <= 0 ) {
					cur_dimens.max_height = dim.max_height;
				} else {
					cur_dimens.max_height =
						max(cur_dimens.max_height,dim.max_height);
				}
				
			// For vertical dimension, similar to above...
			} else {
				min_weight += dim.pref_height-dim.min_height;
				if( i > 0 ) {
					cur_dimens.min_height += intra_s;
					cur_dimens.pref_height += intra_s;
				}
				cur_dimens.min_height += dim.min_height;
				cur_dimens.pref_height += dim.pref_height;
				if( cur_dimens.max_height < ArpAnySize ) {
					cur_dimens.max_height += dim.max_height;
					if( i > 0 ) cur_dimens.max_height += intra_s;
				}
				cur_dimens.min_width =
					max(cur_dimens.min_width,dim.min_width);
				cur_dimens.pref_width =
					max(cur_dimens.pref_width,dim.pref_width);
				if( cur_dimens.max_width <= 0 ) {
					cur_dimens.max_width = dim.max_width;
				} else {
					cur_dimens.max_width =
						max(cur_dimens.max_width,dim.max_width);
				}
			}
			
		}
	}
	
	// Add in space around frame of our group.
	cur_dimens.min_width += inset_l + inset_r;
	cur_dimens.min_height += inset_t + inset_b;
	cur_dimens.pref_width += inset_l + inset_r;
	cur_dimens.pref_height += inset_t + inset_b;
	cur_dimens.max_width += inset_l + inset_r;
	cur_dimens.max_height += inset_t + inset_b;
}

// Scale a number <x> by a factor of <y>/<z>, approximating up.
#define APPROX_SCALE(x,y,z) floor( (z >= 1) ? \
					(((ceil(x))*(y))+((z)-1))/(z) : 0 )

void ArpRunningBar::Layout(void)
{
	// Initialize variables use to spread children across our space.
	BRect frm = LayoutBounds();
	LayoutDimens();
	int32 num = CountLayoutChildren();
	float extra_spc;
	float leftwgt = total_weight;
	float left = frm.left, top = frm.top;
	float width = frm.Width()+1, height = frm.Height()+1;

	// Check if either direction is less than the preferred size.
	bool upw = true, uph = true;
	if( width < cur_dimens.pref_width ) upw = false;
	if( height < cur_dimens.pref_height ) uph = false;
	
	// Move by top/left insets.
	left += inset_l;
	top += inset_t;
	
	BRect cfrm;
	cfrm.left = left;
	cfrm.top = top;
	
	// Spreading across X dimension...
	if( PV_Orientation == B_HORIZONTAL ) {
		// If less then preferred size, our base is min_width and we
		// weight by the difference between preferred width and min width.
		if( !upw ) leftwgt = min_weight;
		// This is the amount of extra space we have to use -- either
		// additional over preferred width, or over the minimum width.
		extra_spc = width -
			(upw ? cur_dimens.pref_width:cur_dimens.min_width);
		// Place the bottom of all children.
		cfrm.bottom = top+height-1-inset_t-inset_b;
	
	// Spreading across Y dimension, similar to above...
	} else {
		if( !uph ) leftwgt = min_weight;
		extra_spc = height -
			(uph ? cur_dimens.pref_height:cur_dimens.min_height);
		cfrm.right = left+width-1-inset_l-inset_r;
	}
	
	ArpD(cdb << ADH << "ArpRunningBar::Layout() -- " << LayoutName()
				<< " = " << frm << endl);
	ArpD(cdb << ADH << "  width=" << width << ", height=" << height
				<< ", extra_spc=" << extra_spc
				<< ", weight=" << leftwgt << endl);

	for( int32 i=0; i<num; i++ ) {
		ArpLayoutable* child = LayoutChildAt(i);
		if( child ) {
			// Retrieve dimensions and constraints of this child.
			const ArpLayoutDimens& dim(child->LayoutDimens());
			float thiswgt = child->Constraints().GetFloat(WeightC,1);
			int32 thisfill = child->Constraints().GetInt32(FillC,ArpCenter);
			
			// If our spread axis is less than its preferred size, the
			// child's weight is instead the difference between its
			// minimum and preferred sizes.
			if( PV_Orientation == B_HORIZONTAL && !upw ) {
				thiswgt = dim.pref_width-dim.min_width;
			} else if( PV_Orientation == B_VERTICAL && !uph ) {
				thiswgt = dim.pref_height-dim.min_height;
			}
			
			// Compute the amount of addition space for this child.
			if( extra_spc < 0 ) extra_spc = 0;
			float spc = APPROX_SCALE(extra_spc,thiswgt,leftwgt);
			
			ArpD(cdb << ADH << "Positioning " << child->LayoutName()
						<< " left=" << left << ", top=" << top
						<< ", extra=" << spc << endl);
			
			// Spreading across X so place left and right of rect...
			if( PV_Orientation == B_HORIZONTAL ) {
				// Currently at left position.
				cfrm.left = left;
				// Set right position to the base size (either min or
				// preferred), plus additional space computed above.
				cfrm.right = floor( left+spc-1 + 
							(upw ? dim.pref_width:dim.min_width) );
				// Set left position for next child.
				left = cfrm.right+intra_s+1;
				
			// Place top and bottom of rect, similarily to above...
			} else {
				cfrm.top = top;
				cfrm.bottom = floor( top+spc-1 + 
							(uph ? dim.pref_height:dim.min_height) );
				top = cfrm.bottom+intra_s+1;
			}
			
			// Compute the final frame of the child, based on its
			// gravity constraint.  If there is more space assigned to
			// the child than it can use, this determines how it is
			// placed within that space.
			BRect ffrm(cfrm);
			const float outw = floor((ffrm.right-ffrm.left+1)-dim.max_width);
			if( outw > 0 ) {
				ArpD(cdb << ADH << "Filling outw=" << outw << ", init frame="
						<< ffrm << endl);
				ArpD(cdb << ADH << "Max width=" << dim.max_width
							<< ", fill mode=" << thisfill << endl);
				switch( thisfill&ArpEastWest ) {
					case ArpCenter: {
						const float leftw = floor(outw/2);
						ffrm.left += leftw;
						ffrm.right -= (outw-leftw);
						break;
					}
					case ArpWest: {
						ffrm.right -= outw;
						break;
					}
					case ArpEast: {
						ffrm.left += outw;
						break;
					}
				}
				ArpD(cdb << ADH << "Final frame=" << ffrm << endl);
			}
			const float outh = floor((ffrm.bottom-ffrm.top+1)-dim.max_height);
			if( outh > 0 ) {
				switch( thisfill&ArpNorthSouth ) {
					case ArpCenter: {
						const float toph = floor(outh/2);
						ffrm.top += toph;
						ffrm.bottom -= (outh-toph);
						break;
					}
					case ArpNorth: {
						ffrm.bottom -= outh;
						break;
					}
					case ArpSouth: {
						ffrm.top += outh;
						break;
					}
				}
			}
			
			// Finally, actually place the child.
			child->SetLayout(ffrm);
			
			// And subtract the space and weight it used.
			extra_spc -= spc;
			leftwgt -= thiswgt;
		}
	}
}
